/*
    aladindump.c

    Converts Aladin raw files produced by aladinlink into UDCF format.

    Copyright (C) 2002 Matthias Cramer, cramer@freestone.net
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  

*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <time.h>
#include "version.h"
#include "config.h"
#include "aladin.h"

extern int errno;
static int checksum(char *);

int main(void) {
  int fd;
  FILE * pf;
  char filename[255];
  char dirname[100];
  char aladinbuf[2048];
  char profilebuf[1536];
  int i,j,k, profilebuflength;
  int rb;
  struct tm * dlt;
  int baseaddress;
  time_t dumptime, temp_time;
  int deltat;
  
  struct s_aladinsettings aladinsettings;
  struct s_aladinstatus aladinstatus;
  struct s_aladinprofile profile[37];
  struct s_aladinlogbook logbook[37];
  struct stat dumpstat;
  struct s_config config;
  t_aladin * aladin;

  printf("aladindump V%s %s\n", VERSION, COPY);
  readconfig(&config);
  
  stat("aladindump.raw", &dumpstat);
  dumptime = dumpstat.st_mtime;

  if (dumpstat.st_size != BUFSIZE) {
    printf("The file 'aladindump.raw' has the wrong size or does not exist!");
    return(1);
  }

  fd = open("aladindump.raw", O_RDONLY);
  rb = read(fd, aladinbuf, BUFSIZE);
  close(fd);
  
  if (rb != BUFSIZE) {
    printf("Error while rading file. File has the frong size.\n");
    return(1);
  }

  if (!checksum(aladinbuf)) {
    return(1);
  }
  
  /* Fill up the settings section */
  
  aladinsettings.model = aladinbuf[0x07bc];
  aladinsettings.is_air = (aladinbuf[0x07bc] &0x0f) % 4 == 0 ? 1 : 0;
  aladinsettings.is_nitrox = (aladinbuf[0x07bc] & 0xf0) == 0xf0 ? 1 : 0;
  aladinsettings.is_o2 = (aladinbuf[0x07bc] & 0xf0) == 0xa0 ? 1 : 0;

  aladinsettings.beep  = (aladinbuf[0x07d2] & 0x02) == 0x02 ? 1 : 0;
  aladinsettings.units = (aladinbuf[0x07d2] & 0x01) == 0x01 ? 1 : 0;

  aladinsettings.reserve = aladinbuf[0x07de];

  aladinsettings.brethwarning = aladinbuf[0x07eb];

  aladinsettings.serial = (aladinbuf[0x07ed] << 16) + 
                          (aladinbuf[0x07ee] << 8) + 
                          aladinbuf[0x07ef];

  /*  Fill up status section */
  
  aladinstatus.battery = (unsigned char) aladinbuf[0x07f0] * 100 / 256;
  aladinstatus.batterystatus = (aladinbuf[0x07f1] & 0x0f) == 0 ? 1 : 0;
  
  aladinstatus.totaldives = ((unsigned char) aladinbuf[0x07f2] << 8) + 
                            (unsigned char) aladinbuf[0x07f3];
  aladinstatus.logbookoffset = ((aladinbuf[0x07f4] +36) % 37) * 12 +0x0600;
  aladinstatus.numprofiles = aladinbuf[0x07f5];
  aladinstatus.endringbuf = (((aladinbuf[0x07f7] << 7) & 0xff00) | aladinbuf[0x07f6]) & 0x07ff;
  aladinstatus.time = ((aladinbuf[0x07f8] << 24) + 
                            (aladinbuf[0x07f9] << 16) + 
                            (aladinbuf[0x07fa] << 8) + 
                            aladinbuf[0x07fb]) / 2 + ALADINEPOCH;


  /* and now to the logbook */
  
  i = aladinstatus.numprofiles;
  
  for (baseaddress = aladinstatus.logbookoffset; i > 0; baseaddress -= 12) {
    if (baseaddress < 0x0600) {
      baseaddress = 0x0600;
    }

    logbook[i].divenum = aladinstatus.totaldives - aladinstatus.numprofiles + i;
    logbook[i].altitude = ( aladinbuf[baseaddress] & 0xb0 ) >> 6;
    logbook[i].sos = ( aladinbuf[baseaddress] & 0x20 ) >> 5;
    logbook[i].workalarm = ( aladinbuf[baseaddress] & 0x10 ) >> 4;
    logbook[i].decoalarm = ( aladinbuf[baseaddress] & 0x08 ) >> 3;
    logbook[i].repeatd = ( aladinbuf[baseaddress] & 0x02 ) >> 1;
    logbook[i].acentwaring = aladinbuf[baseaddress] & 0x01;

    logbook[i].bottomtime = ((aladinbuf[baseaddress] & 0x04 ) >> 2) * 100 +
                            ((aladinbuf[baseaddress+1] & 0xf0) >> 4) * 10 + 
                            (aladinbuf[baseaddress+1] & 0x0f);

    /* Formula that strips out garbage in the last 6 bits */
    logbook[i].maxdepth = (((aladinbuf[baseaddress+2] << 8) + aladinbuf[baseaddress+3]) >>6) * 10.0 /64.0;

    /* Formula used as in ROM Program and Datatrak */
    /* 
      logbook[i].maxdepth = (((aladinbuf[baseaddress+2] << 8) + aladinbuf[baseaddress+3])) * 10.0 /4096.0;
    */
    
    if (logbook[i].repeatd) {
      logbook[i].surfaceh = ((aladinbuf[baseaddress+4] & 0xf0) >> 4) * 10 +
                            (aladinbuf[baseaddress+4] & 0x0f);          
                          
      logbook[i].surfacem = ((aladinbuf[baseaddress+5] & 0xf0) >> 4) * 10 +
                           (aladinbuf[baseaddress+5] & 0x0f);          
    } else {
      logbook[i].surfaceh = 0;
      logbook[i].surfacem = 0;
    }

    logbook[i].airconsumption = aladinbuf[baseaddress+6] * 1.378951458634;
    
    logbook[i].entrytime = ((aladinbuf[baseaddress+7] << 24) + 
                           (aladinbuf[baseaddress+8] << 16) + 
                           (aladinbuf[baseaddress+9] << 8) + 
                           aladinbuf[baseaddress+10]) / 2 + ALADINEPOCH;
    
    logbook[i].watertemp = ((unsigned char) aladinbuf[baseaddress+11]) / 4.0;
    
    printf("------------------------ Dive %d --------------------------------\n", logbook[i].divenum);
    printf("Altitude level    : %d\n", logbook[i].altitude);
    printf("Alarms            : SOS %d, WORK %d, DECO %d, ACENT %d\n", logbook[i].sos, logbook[i].workalarm, logbook[i].decoalarm, logbook[i].acentwaring);
    printf("Bottomtime        : %d\n", logbook[i].bottomtime);
    printf("Max. Depth        : %.2f\n", logbook[i].maxdepth);
    printf("Surface Interval  : %02d:%02d\n", logbook[i].surfaceh, logbook[i].surfacem);
    printf("Air consunption   : %.2f Bar\n", logbook[i].airconsumption);
    dlt = localtime(&(logbook[i].entrytime));
    printf("Entry Time        : %s", asctime(dlt));
    printf("Water Temperature : %.2f\n", logbook[i].watertemp);
    
    i--;
  }
  
  /* and last, the dive profiles */
  
  /* first find the oldest logbook entry */
  for (baseaddress = aladinstatus.endringbuf; (unsigned char) aladinbuf[baseaddress] != 0xff; baseaddress++) {
    if (baseaddress > 0x5ff) {
      baseaddress = -1;
    }   
  }

  printf("Start of oldest diveprofile at: %#x, value at that position: %#x\n", baseaddress, (unsigned char) aladinbuf[baseaddress]);
  
  /* Copy profiles to a linear buffer */
  for (i=0; baseaddress+i != aladinstatus.endringbuf; i++) {
    if (baseaddress+i > 0x5ff) {
//      baseaddress = -1;
      baseaddress = 0;
      i=0;
    }
    profilebuf[i] = aladinbuf[baseaddress+i];
  }
  profilebuf[i] = aladinbuf[baseaddress+i];
  
  profilebuflength = i;
  
  baseaddress = 0;

  for (i=1; i <= aladinstatus.numprofiles; i++) {
    
    profile[i].divenum = aladinstatus.totaldives - aladinstatus.numprofiles + i;
    profile[i].ambienttemp = ((unsigned char) profilebuf[baseaddress+1]) / 4.0;

/*
    printf("Dive Number        : %d\n", profile[i].divenum);
    printf("Ambient Temperatur : %.2f\n", profile[i].ambienttemp);
*/
    
    profile[i].tissue1 = profilebuf[baseaddress+2] + (profilebuf[baseaddress+2] << 8);
    profile[i].tissue2 = profilebuf[baseaddress+4] + (profilebuf[baseaddress+5] << 8);
    profile[i].tissue3 = profilebuf[baseaddress+6] + (profilebuf[baseaddress+7] << 8);
    profile[i].tissue4 = profilebuf[baseaddress+8] + (profilebuf[baseaddress+9] << 8);
    profile[i].tissue5 = profilebuf[baseaddress+10] + (profilebuf[baseaddress+11] << 8);
    profile[i].tissue6 = profilebuf[baseaddress+12] + (profilebuf[baseaddress+13] << 8);
    profile[i].tissue7 = profilebuf[baseaddress+14] + (profilebuf[baseaddress+15] << 8);
    profile[i].tissue8 = profilebuf[baseaddress+16] + (profilebuf[baseaddress+17] << 8);
    
    profile[i].microbubbles = profilebuf[baseaddress+18] + ((profilebuf[baseaddress+19] & 0xf0 ) << 4);
    profile[i].intrapulmenary = ((profilebuf[baseaddress+19] & 0x0f ) << 8) + profilebuf[baseaddress+20];

    profile[i].skincool = (unsigned int) (((profilebuf[baseaddress+21] + (profilebuf[baseaddress+22] & 0xf0)) << 4) >> 6);

    /* O2 and nitrox stiff */

    baseaddress+=22;
    
/*    
    printf("profilebuf[baseaddress+22] : %#x [%#x -> %d]\n", (unsigned char) profilebuf[baseaddress], baseaddress, baseaddress);
*/

    if (aladinsettings.is_o2) {
      baseaddress += 3;
      printf("is_02\n");
    } else if (aladinsettings.is_nitrox) {
      baseaddress += 2;
      printf("is_nitrox\n");
    }
    
    j = 0;
    k = 0;
    
    baseaddress++;
    
    while ((unsigned char) profilebuf[baseaddress] != 0xff) {
      profile[i].diveprofile[j] = (((profilebuf[baseaddress] << 8) + profilebuf[baseaddress+1]) >> 6) * 0.15625;
      
      profile[i].warnings[j].transmit = (profilebuf[baseaddress+1] & 0x20) >> 5;
      profile[i].warnings[j].work = (profilebuf[baseaddress+1] & 0x10) >> 4;
      profile[i].warnings[j].decomissed = (profilebuf[baseaddress+1] & 0x08) >> 3;
      profile[i].warnings[j].ascent = (profilebuf[baseaddress+1] & 0x04) >> 2;
      profile[i].warnings[j].bottom = (profilebuf[baseaddress+1] & 0x02) >> 1;
      profile[i].warnings[j].decostop = profilebuf[baseaddress+1] & 0x01;
      
      baseaddress += 2;

      if (((unsigned char) profilebuf[baseaddress] != 0xff) && (j % 3 == 2)) {
      
        /* printf("i j k = %d %d %d\n", i, j, k); */
        
        profile[i].decoinfo[k].physeffort = (profilebuf[baseaddress] & 0x70) >> 4;

        /* 
        Have to look at this ...
          profile[i].decoinfo[k].skincooling;
        */
      
        profile[i].decoinfo[k].microbubble = profilebuf[baseaddress] & 0x07;
      
        baseaddress++;
        k++;
      }
      
      if (aladinsettings.is_o2) {
        printf("is_02\n");
        baseaddress++;
      }
      j++;
    }
    profile[i].profilesamples = j;  
  }

  /* display the values form the settings section */

  for (aladin = aladintype; aladin->name; aladin++) {
    if (aladinsettings.model == aladin->type) {
      break;
    }
  }

  printf("\n");
  printf("Model: %s, Serial: %u\n", aladin->name, aladinsettings.serial);
  printf("IS_AIR: %x, IS_NITROX: %x, IS_O2: %x\n", aladinsettings.is_air, aladinsettings.is_nitrox, aladinsettings.is_o2);
  printf("Beep: %s, Unit: %s\n", aladinsettings.beep == 1 ? "ON" : "OFF", aladinsettings.units == 0 ? "Metric" : "Imperial");
  printf("Reserve: %d Bar\n", aladinsettings.reserve);
  printf("Brethwarning: %d\n", aladinsettings.brethwarning);
  
  printf("Battery: %u %% %s\n", aladinstatus.battery, aladinstatus.batterystatus == 1 ? "OK" : "OFF");
  printf("Total Dives: (%u + %u) %u\n", aladinstatus.totaldives, config.aladindiveoffset, config.aladindiveoffset + aladinstatus.totaldives);
  
  printf("Newest Logbook Entry at: %#x (%u), End of Buffer: %#x\n", aladinstatus.logbookoffset, (unsigned char) aladinbuf[0x07f4], aladinstatus.endringbuf);
  
  deltat = dumptime - aladinstatus.time;
  
  aladinstatus.time += deltat;
  
  dlt = localtime(&(aladinstatus.time));
  
  printf("Data was downloaded on : %s\n", asctime(dlt));

  /* write the data out in UDCF format */

  for(i=1; i <= aladinstatus.numprofiles; i++) {
  
    /* Look if ".eon" does exist */
    sprintf(dirname, "%s/.aladin",getenv("HOME"));
    
    stat(dirname, &dumpstat);
    
    if ((dumpstat.st_mode & S_IFDIR) != S_IFDIR) {
      printf("Create .aladin directory ...");
      mkdir(dirname, 0755);
      printf(" done\n");
    }   

    temp_time = logbook[i].entrytime + deltat;

    dlt = localtime(&temp_time);

    sprintf(filename, "%s/%d-%04d-%02d-%02d_%02d%02d.udcf", dirname, logbook[i].divenum+config.aladindiveoffset, dlt->tm_year+1900, dlt->tm_mon+1, dlt->tm_mday, dlt->tm_hour, dlt->tm_min);
    if (open(filename,O_RDONLY) == -1) {
      pf = fopen(filename,"w");
      if (pf == NULL) {
        fprintf(stderr, "Error: Can not open UDCF file!\n");
        return(1);
      }
      printf("Dive will be saved in UDCF format as %s\n", filename);
      
      fprintf(pf, "<PROFILE UDCF=\"1\">\n");
      fprintf(pf, "<UNITS>Metric</UNITS>\n");
      fprintf(pf, "<DEVICE>\n <VENDOR>UWATEC</VENDOR>\n <MODEL>%s</MODEL>\n <VERSION>1</VERSION>\n</DEVICE>\n", aladin->name);
      fprintf(pf, "<ADDINFO>\n");
      fprintf(pf, " <PERSINFO>%s</PERSINFO>\n", config.aladinowner);
      fprintf(pf, " <TOTALDIVES>%d</TOTALDIVES>\n", logbook[i].divenum+config.aladindiveoffset);
      fprintf(pf, " <SERIALID>%d</SERIALID>\n", aladinsettings.serial);   
      fprintf(pf, "</ADDINFO>\n");
      fprintf(pf, "<REPGROUP>\n");
      fprintf(pf, " <DIVE>\n");
      fprintf(pf, "  <DATE><DAY>%d</DAY><MONTH>%d</MONTH><YEAR>%04d</YEAR></DATE>\n", dlt->tm_mday, dlt->tm_mon+1, dlt->tm_year+1900);
      fprintf(pf, "  <TIME><HOUR>%d</HOUR><MINUTE>%d</MINUTE></TIME>\n", dlt->tm_hour, dlt->tm_min);
      fprintf(pf, "  <SURFACEINTERVAL>%d</SURFACEINTERVAL>\n", logbook[i].surfaceh * 60 + logbook[i].surfacem);
      fprintf(pf, "  <TEMPERATURE>%.2f</TEMPERATURE>\n", logbook[i].watertemp);
      fprintf(pf, "  <DENSITY>1030.000</DENSITY>\n");
      fprintf(pf, "  <ALTITUDE>%d</ALTITUDE>\n", aladinaltitudes[logbook[i].altitude]);
      fprintf(pf, "  <GASES>\n");                            
      fprintf(pf, "   <MIX>\n");
      fprintf(pf, "    <MIXNAME>RealAir</MIXNAME>\n");   
      fprintf(pf, "    <TANK><PDELTA>%.0f</PDELTA></TANK>\n", logbook[i].airconsumption * 100000);
//      fprintf(pf, "    <TANKVOLUME>15.0</TANKVOLUME>\n");
//      fprintf(pf, "    <TANK><PSTART>%d00000.0</PSTART><PEND>%d00000.0</PEND></TANK>\n", dive[i].startpresure, dive[i].endpresure);
      fprintf(pf, "    <O2>0.21</O2>\n");
      fprintf(pf, "    <N2>0.79</N2>\n");
      fprintf(pf, "    <HE>0.00</HE>\n");
      fprintf(pf, "   </MIX>\n");        
      fprintf(pf, "  </GASES>\n");
      fprintf(pf, "  <DELTAMODE/>\n");
      fprintf(pf, "  <SAMPLES>\n");
      fprintf(pf, "   <DELTA>%d</DELTA>\n", 20);
      fprintf(pf, "   <SWITCH>RealAir</SWITCH>\n");
      
      fprintf(pf, "   <D>0.0</D>\n");


//      k = (logbook[i].bottomtime * 3) < profile[i].profilesamples ? logbook[i].bottomtime * 3 : profile[i].profilesamples;
//      for (j=0; j<=k; j++) {

      for (j=0; j<=profile[i].profilesamples; j++) {

        if (profile[i].warnings[j].decostop) {
          fprintf(pf, "   <ALARM>DECO</ALARM>\n");
        }

        if (profile[i].warnings[j].work) {
          fprintf(pf, "   <ALARM>WORK</ALARM>\n");
        }

        if (profile[i].warnings[j].decomissed) {
          fprintf(pf, "   <ALARM>DECO MISSED</ALARM>\n");
        }

        if (profile[i].warnings[j].ascent) {
          fprintf(pf, "   <ALARM>ASCENT</ALARM>\n");
        }

        if (profile[i].warnings[j].bottom) {
          fprintf(pf, "   <ALARM>BOTTOMTIME</ALARM>\n");
        }

        if (profile[i].warnings[j].transmit) {
          fprintf(pf, "   <ALARM>TRANSMIT</ALARM>\n");
        }
       
        fprintf(pf, "   <D>%6.2f</D>\n", profile[i].diveprofile[j]);

      }
      
      fprintf(pf, "  </SAMPLES>\n");
      fprintf(pf, " </DIVE>\n");
      fprintf(pf, "</REPGROUP>\n");
      fprintf(pf, "</PROFILE>\n");

      fclose(pf);
    } else {
      printf("Dive '%s' already exists, skipping [%s (%d)] ...\n", filename, strerror(errno), errno);
    }
  }

  return(0);  
}  /* end main */


static int checksum(char * buf) {
  int i;
  unsigned int csum = 0;
  unsigned int filecsum;
  
  filecsum = (((unsigned char) buf[0x07fd] << 8) | (unsigned char) buf[0x07fc]);
  /* filecsum &= 0x0ffff; */

  printf("File Checksum %#04x ... ", filecsum);

  for (i=0; i < 0x07fc; i++) {
    csum += (unsigned char) buf[i];
  }

  csum += 0x1fe;
  csum %= 65536;
  csum &= 0x0ffff;

  if ( csum == filecsum) {
    printf("OK\n");
    return(1);
  } else {
    printf("Error\n");
    return(0);
  }
}

